;; Generated package description from vc-jj.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "vc-jj" "0.2" "VC backend for the Jujutsu version control system" '((emacs "28.1") (compat "29.4")) :commit "91c15146d14c377cd97d191fd515d1ca8eb91427" :authors '(("Rudolf Schlatte" . "rudi@constantly.at")) :maintainer '("Rudolf Schlatte" . "rudi@constantly.at") :keywords '("vc" "tools") :url "https://codeberg.org/emacs-jj-vc/vc-jj.el")
